/*	SaveRelocator
	by MK2k
	
	Notes:	- R6 saves end with .nds.SAV
	
	Released under GPLv2 and GPLv3
*/

#include <PA9.h>
#include <fat.h>
#include <sys/dir.h>
#include <memory.h>
#include <stdarg.h>
#include <stdio.h>

struct SaveConnection
{
	char cFullPath[256];
	char cPath[256];
	char cFileName[256];
	bool bDone;
	
	struct SaveConnection *next;
};

struct SaveConnection *scResults[100];	// Resultset, contains all locations of a specific savefile

void logprint(char *message)
{
	FILE *log = fopen("/log.txt", "at");
    if (!log) log = fopen("log.txt", "wt");
    if (!log)
	{
        return;   // bail out if we can't log
    }

	fprintf(log, "%s", message);

    fclose(log);
}

void copymove(char *cOldFile, char *cNewFile)
{
	FILE *out, *in;
	out = fopen(cOldFile, "rb");
	if(!out)
		return;
	
	in = fopen(cNewFile, "wb");
	if(!in)
	{
		fclose(out);
		return;
	}
	
	int len;
	char buffer[1024];
	
	while((len = fread(buffer,1,sizeof(buffer),out)) > 0)
	{
		fwrite(buffer,1,len,in);
	}
	
	fclose(in);
	fclose(out);
	
	remove(cOldFile);
	

	return;
}

struct SaveConnection *createSaveConnection(char *cASaveFile, char *cADirectory, struct SaveConnection *root)
{
	if(cASaveFile && cADirectory)
	{
		//TODO: check for file existence
		
		struct SaveConnection *newConn = malloc(sizeof(struct SaveConnection));

		struct SaveConnection *last = NULL;
		if(root)
		{
			last = root;
			
			while(last->next)
			{
				last = last->next;
			}
		}

		sprintf(newConn->cFullPath, "%s/%s", cADirectory, cASaveFile); //TODO: distinguish!
		sprintf(newConn->cPath, "%s", cADirectory);
		sprintf(newConn->cFileName, "%s", cASaveFile);
		newConn->bDone = false;
		
		if(last)
		{
			last->next = newConn;
		}
		
		return newConn;
	}

	return NULL;
}

bool checkOriginalFile(char *cFileName, char *cDir)
{
	if(cFileName && cDir)
	{
		char cOrigName[256];
		strncpy(cOrigName, cFileName, strlen(cFileName) - 4); // copy the save's filename over but without the '.sav'
		cOrigName[strlen(cFileName) - 4] = 0;	// zero termination
		
		DIR_ITER *dir = diropen(cDir);
		struct stat st;
		char filename[256];
		while (dirnext(dir, filename, &st) == 0)
		{
			if(!(st.st_mode & S_IFDIR))
			{
				if(strstr(filename, cOrigName) && !strstr(filename, ".nds.SAV"))
				{
					return true; // found the original
				}
			}
		}
	}
	
	return false;  // didnt find the original
}

// Scan through all Directories recursively and...
// find the local .nds.SAV files
// if a corresponding .nds file does not exist there, add the fullpath to the *root's last node
struct SaveConnection *scanDirs(char *cDir, struct SaveConnection *root)
{
	if(!cDir)
		return root;
	
	DIR_ITER *dir = diropen(cDir);
	struct stat st;
	char filename[256];
	
	if (dir == NULL)
		return root;
	
	while (dirnext(dir, filename, &st) == 0)
	{
		if(st.st_mode & S_IFDIR)
		{
			// Directory found!
			if(strcmp(filename, ".") && strcmp(filename, ".."))
			{
				char cTemp[256];
				if(cDir[strlen(cDir)-1] == '/')
				{
					sprintf(cTemp, "%s%s", cDir, filename);
				}
				else
				{
					sprintf(cTemp, "%s/%s", cDir, filename);
				}
				root = scanDirs(cTemp, root);
			}
		}
		else
		{
			// Not a Directory, check if it's a .nds.SAV file
			if(strstr(filename, "nds.SAV"))
			{
				//found a "nds.SAV" file
				if(!checkOriginalFile(filename, cDir))
				{
					// the original file isn't in this directory -> add the file to the linked list
					if(!root)
					{
						root = createSaveConnection(filename, cDir, NULL);
					}
					else
					{
						createSaveConnection(filename, cDir, root);
					}
				}
			}
		}
	}
	return root;
}

int buildResultSet(char *cSaveFile, char *cOrigPath, struct SaveConnection *root)
{
	int i;
	for(i=0; i < 100; i++)
	{
		scResults[i] = NULL;
	}
	
	int iResult = 0;
	struct SaveConnection *ptr = root;
	while(ptr)
	{
		if(!strcmp(ptr->cFileName, cSaveFile))
		{
			// new instance found!
			scResults[iResult] = ptr;
			iResult++;
		}
	
		ptr = ptr->next;
	}
	
	return iResult;
}

void handleResult(int iNumRes, char *cDir, char *filename, struct SaveConnection *ptr)
{
	if (iNumRes <= 0)
	{
		return;
	}
	
	int iSaveExists = 0;
	char newfile[256];
	if(cDir[strlen(cDir)-1] == '/')
	{
		sprintf(newfile, "%s%s.SAV", cDir, filename);
	}
	else
	{
		sprintf(newfile, "%s/%s.SAV", cDir, filename);
	}

	FILE *fTemp = fopen(newfile, "r");
	if(fTemp)
	{
		// There already exists a Savefile in the target location
		fclose(fTemp);
		iSaveExists = 1;
	}
	
	//Handle only one occurence first!
	if(iNumRes + iSaveExists == 1)
	{
		consoleClear();
		printf("%i/%i occurence of\n", 1, 1);
		printf("%s\n", filename);
		printf("in: %s\n", scResults[0]->cPath);
		printf("--------------------------------\n");
		printf("Moving...");
		copymove(scResults[0]->cFullPath, newfile);
		scResults[0]->bDone = true;
		printf("Done\n");
		PA_WaitForVBL();
		return;
	}
	
	bool bInMenu = true;
	int iCurrentSave = 0;
	while(bInMenu)
	{
		consoleClear();
		printf("%i/%i occurence of\n", iCurrentSave+1, iNumRes + iSaveExists);
		printf("%s\n", filename);
		if(iCurrentSave < iNumRes)
		{
			printf("in: %s\n", scResults[iCurrentSave]->cPath);
		}
		else
		{
			printf("in: %s\n", cDir);
			printf("(this is the game's folder!)\n");
		}
		
		printf("--------------------------------\n");
		printf("A: Use this, keep the others\n");
		printf("X: Use this, delete the others\n");
		printf("Left: Previous occurence\n");
		printf("Right: Next occurence\n");
		
		bool bWaitInput = true;
		while(bWaitInput)
		{
			if (Pad.Newpress.Left)
			{
				//previous
				iCurrentSave--;
				if(iCurrentSave < 0)
				{
					iCurrentSave = iNumRes - 1 + iSaveExists;
					bWaitInput = false;
				}
				bWaitInput = false;
			}
			if (Pad.Newpress.Right)
			{
				// next
				iCurrentSave++;
				if(iCurrentSave > iNumRes - 1 + iSaveExists)
				{
					iCurrentSave = 0;
				}
				bWaitInput = false;
			}
			if (Pad.Newpress.A)
			{
				// copymove (does nothing if it's the current save file)
				if(iCurrentSave < iNumRes)
				{
					printf("Moving...");
					copymove(scResults[iCurrentSave]->cFullPath, newfile);
					scResults[iCurrentSave]->bDone = true;
				}
				int i;
				for(i=0; i < iNumRes; i++)
				{
					scResults[i]->bDone = true;
				}
				printf("Done\n");
				bWaitInput = false;
				bInMenu = false;
			}
			if (Pad.Newpress.X)
			{
				// copymove (does nothing if it's the current save file)
				if(iCurrentSave < iNumRes)
				{
					printf("Moving...");
					copymove(scResults[iCurrentSave]->cFullPath, newfile);
					scResults[iCurrentSave]->bDone = true;
				}
				
				// delete others
				if(iNumRes > 1)
				{
					printf("Deleting...");
				}
				int i;
				for(i = 0; i < iNumRes; i++)
				{
					if(i != iCurrentSave)
					{
						remove(scResults[i]->cFullPath);
						scResults[i]->bDone = true;
					}
				}
				printf("Done\n");
				bWaitInput = false;
				bInMenu = false;
			}
			PA_WaitForVBL();
		}
	}
}

void relocateSaves(char *cDir, struct SaveConnection *root)
{
	if(!cDir)
		return;
	
	DIR_ITER *dir = diropen(cDir);
	struct stat st;
	char filename[256];
	
	if (dir == NULL)
		return;
	
	while (dirnext(dir, filename, &st) == 0)
	{
		if(st.st_mode & S_IFDIR)
		{
			// Directory found!
			if(strcmp(filename, ".") && strcmp(filename, ".."))
			{
				char cTemp[256];
				if(cDir[strlen(cDir)-1] == '/')
				{
					sprintf(cTemp, "%s%s", cDir, filename);
				}
				else
				{
					sprintf(cTemp, "%s/%s", cDir, filename);
				}
				relocateSaves(cTemp, root);
			}
		}
		else
		{
			// Not a Directory, check if it's a .nds.SAV file
			if(!strstr(filename, "nds.SAV") && strstr(filename, ".nds"))
			{
				//found a ".nds" file, now searching through our found saves and if one is matching, renaming it over
				if(root)
				{
					struct SaveConnection *ptr = root;
					while(ptr)
					{
						if(!ptr->bDone)
						{
							if(strstr(ptr->cFullPath, filename))
							{
								// We found an original file for our save file!
								char newfile[256];
								
								if(cDir[strlen(cDir)-1] == '/')
								{
									sprintf(newfile, "%s%s.SAV", cDir, filename);
								}
								else
								{
									sprintf(newfile, "%s/%s.SAV", cDir, filename);
								}
								int iNumRes = buildResultSet(ptr->cFileName, ptr->cPath, root);
								handleResult(iNumRes, cDir, filename, ptr);
							}
						}
						ptr = ptr->next;
					}
				}

			}
		}
	}
}

int main(int argc, char ** argv)
{
	PA_Init();
	PA_InitVBL();
	
	PA_WaitForVBL();  PA_WaitForVBL();  PA_WaitForVBL();  // wait a few VBLs
	
	consoleDemoInit();
	
	fatInitDefault();

	printf("SaveRelocator 1.0 by MK2k\n-------------------------\n");
	printf("Scanning for .SAV files...\n");

	struct SaveConnection *root = NULL;
	root = scanDirs("/", root);

	// Now each found savestate is in root and it's linked elements
	// Scan again over all directories, if we find corresponding original files, rename the saves over to their correct location
	if(root)
	{
		relocateSaves("/", root);
		consoleClear();
	}
	else
	{
		printf("No deserted Savefile\nhas been found.\n");
	}

	printf("\nThat's all folks!\nVisit www.mk2k.net");
	
	//nothing to do but wait....
	int iRun = 1;
	while (iRun)
	{ 
		PA_WaitForVBL();
	}
	
	return 0;
}
